import os
import re
import aiohttp
import asyncio
from urllib.parse import urlparse
from os.path import join as pjoin
import xml.etree.ElementTree as ET

from tqdm.asyncio import tqdm_asyncio
import requests


def find_relative_path(url, index_url):
    """
    Finds the relative path of a file URL with respect to an index URL.

    :param url: The full URL to the file.
    :param index_url: The base URL representing the root directory.
    :return: The relative path from the index URL to the file's directory.
    """
    # Parse the URLs to extract path components
    url_path = urlparse(url).path
    index_url_path = urlparse(index_url).path
    # Ensure the base path ends with a slash
    if not index_url_path.endswith("/"):
        index_url_path += "/"
    # Construct the full path for comparison and remove the base URL path
    relative_full_path = url_path.replace(index_url_path, "", 1)
    # Remove the filename from the path to get the directory path
    relative_dir_path = "/".join(relative_full_path.split("/")[:-1])
    return relative_dir_path


async def download_file(url, dest_dir):
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
    local_filename = os.path.join(dest_dir, url.split("/")[-1])

    tout = aiohttp.ClientTimeout(10**10)
    async with aiohttp.ClientSession(timeout=tout) as session:
        async with session.get(url) as resp:
            total_size = int(resp.headers.get("content-length", 0))
            chunk_size = 1024
            with tqdm_asyncio(
                total=total_size,
                unit="B",
                unit_scale=True,
                desc=local_filename,
                ascii=True,
            ) as progress_bar:
                with open(local_filename, "wb") as f:
                    async for chunk in resp.content.iter_chunked(chunk_size):
                        f.write(chunk)
                        progress_bar.update(len(chunk))
    return local_filename


async def download_files(urls, dest_dir, n_parallel=5, base_url=None):
    semaphore = asyncio.Semaphore(n_parallel)

    async def bounded_download(url):
        async with semaphore:
            if base_url:
                r_path = find_relative_path(url, base_url)
                dest_dir_ = pjoin(dest_dir, r_path)
            else:
                dest_dir_ = dest_dir
            await download_file(url, dest_dir_)

    await asyncio.gather(*[bounded_download(url) for url in urls])


def download_content(url: str) -> str:
    """
    Downloads the content at the specified URL and returns it as a string.

    :param url: The URL of the content to download.
    :return: The content of the URL as a string.
    """
    response = requests.get(url)
    response.raise_for_status()
    return response.text


def parse_s3_xml(content: str, key_pattern: str) -> list[dict]:
    # Parse the XML content
    root = ET.fromstring(content)

    # Define the namespace mapping
    ns = {"ns": "http://s3.amazonaws.com/doc/2006-03-01/"}

    # Compile the regex pattern for matching keys
    pattern = re.compile(key_pattern)

    # Find all 'Contents' elements considering the namespace
    contents_elements = root.findall("ns:Contents", ns)

    # Extract the 'Key' element text if it matches the pattern
    items = []
    for elem in contents_elements:
        key = elem.find("ns:Key", ns)
        if key is None:
            continue
        key_text = key.text
        if key_text and pattern.match(key_text):
            size_field = elem.find("ns:Size", ns)
            if (size_field is None) or (size_field.text is None):
                continue
            size = int(size_field.text)
            items.append({
                "key": key_text,
                "size": size,
            })
    return items


if __name__ == "__main__":
    import fire

    fire.Fire(download_files)




S3_BASE_URL = "https://uk1s3.embassy.ebi.ac.uk"
S3_MODELS_URL = f"{S3_BASE_URL}/model-repository/"
S3_CONDA_ENVS_URL = f"{S3_BASE_URL}/bioimage.io/environments"
TRITON_IMAGE = "docker://nvcr.io/nvidia/tritonserver:23.03-py3"
S3_IMAGE = "docker://minio/minio:RELEASE.2023-10-24T04-42-36Z.hotfix.793d9ee53"

DOCKER_REGISTRY_PREFIX = "docker://ghcr.io/bioimage-io"

IMJOY_APP_TEMPLATE = """
{app_code}

async def main():
    services = []
    server = api
    _register_service = server.register_service
    async def patched_register_service(*args, **kwargs):
        svc_info = await _register_service(*args, **kwargs)
        services.append(svc_info)
        return svc_info
    server.register_service = patched_register_service
    server.registerService = patched_register_service

    try:
        client_id = server.rpc.get_client_info()['id']
        await hypha_startup(server)
        report_service = await server.get_service("hypha-launcher")
        await report_service.report_services(client_id, services)
    except Exception as e:
        print("Failed to start hypha server:", e)
        await report_service.report_error(client_id, str(e))

import asyncio
loop = asyncio.get_event_loop()
loop.create_task(main)
loop.run_forever()
"""


REPORT_ADDRESS_SCRIPT = """
import json
import http.client
from executor.engine.utils import PortManager

http_port = PortManager.get_port()
task_id = "{task_uuid}"
post_data = {{"port": http_port, "uuid": task_id}}
post_data_json = json.dumps(post_data)
for ip in {host_ips}:
    try:
        conn = http.client.HTTPConnection(ip, {ip_record_server_port})
        conn.request("POST", "/", post_data_json, {{"Content-Type": "application/json"}})
        response = conn.getresponse()
        if response.status == 200:
            break
    except Exception as e:
        print("Failed to record IP:", e)
else:
    raise ValueError("Failed to record IP")

"""

LAUNCH_TRITON_SCRIPT = REPORT_ADDRESS_SCRIPT + f"""
import os
from hypha_launcher.utils.container import ContainerEngine

container_engine_kwargs = {{container_engine_kwargs}}
container_engine = ContainerEngine(**container_engine_kwargs)
container_engine.pull_image("{TRITON_IMAGE}")

models_dir = "{{model_repository}}"
volumes = dict([(models_dir, "/models")])
ports = dict([(http_port, 8000)])
envs = dict([("TRITON_SERVER_URL", f"127.0.0.1:" + str(http_port))])

triton_cmd = 'bash -c "tritonserver --model-repository=/models --log-verbose=3 --log-info=1 --log-warning=1 --log-error=1 --model-control-mode=poll --exit-on-error=false --repository-poll-secs=10 --allow-grpc=False --http-port=' + str(http_port) + '"'
cmd = container_engine.get_command(
    triton_cmd, "{TRITON_IMAGE}",
    volumes=volumes,
    envs=envs,
)
print(cmd)
os.system(cmd)
"""



import os
import typing as T
from pathlib import Path
import secrets
import uuid
import asyncio

from executor.engine import Engine
from executor.engine.job.extend import SubprocessJob, WebappJob
from executor.engine.job import Job, ProcessJob
from executor.engine.utils import PortManager
from imjoy_rpc.hypha import connect_to_server
from pyotritonclient import get_config, execute

from .utils.download import download_files, parse_s3_xml, download_content
from .utils.log import get_logger
from .utils.container import ContainerEngine
from .utils.hpc import HPCManger
from .utils.misc import get_all_ips
from .constants import (
    S3_MODELS_URL, S3_CONDA_ENVS_URL, LAUNCH_TRITON_SCRIPT,
    S3_IMAGE, TRITON_IMAGE
)


logger = get_logger()


class HyphaLauncher:
    def __init__(
            self,
            store_dir: str = ".hypha_launcher",
            debug: bool = False,
            container_engine_kwargs: T.Optional[T.Dict[str, T.Any]] = None,
            hpc_manager_kwargs: T.Optional[T.Dict[str, T.Any]] = None,
            executor_engine_kwargs: T.Optional[T.Dict[str, T.Any]] = None,
            ):
        store_dir = os.environ.get("HYPHA_LAUNCHER_STORE_DIR", store_dir)
        self.store_dir = Path(store_dir).expanduser().absolute()
        if not self.store_dir.exists():
            self.store_dir.mkdir(parents=True)
        logger.info(f"Store dir: {self.store_dir}")
        self.debug = debug
        if container_engine_kwargs is None:
            container_engine_kwargs = {
                "store_dir": str(self.store_dir / "containers"),
            }
        self._container_engine_kwargs = container_engine_kwargs
        self.container_engine = ContainerEngine(**container_engine_kwargs)
        if hpc_manager_kwargs is None:
            hpc_manager_kwargs = {}
        self.hpc_manager = HPCManger(**hpc_manager_kwargs)
        if executor_engine_kwargs is None:
            executor_engine_kwargs = {}
        self.engine = Engine(**executor_engine_kwargs)
        self._task_uuid_to_job: T.Dict[str, Job] = {}
        self._ip_record_server_job: T.Optional[Job] = None
        self._ip_record_flie = self.store_dir / "tmp" / "record_ip.txt"

    def get_free_port(self) -> int:
        return PortManager.get_port()

    def get_jobs_ids(self) -> T.List[str]:
        return [job.id for job in self.engine.jobs]

    async def stop_job(self, job_id: str) -> bool:
        job = self.engine.jobs.get_job_by_id(job_id)
        if job is not None:
            await job.cancel()
            return True
        return False

    async def download_from_s3(
        self,
        pattern: str,
        dest_dir: T.Optional[str] = None,
        n_parallel: int = 5,
        s3_base_url=S3_MODELS_URL,
    ):
        """Download files from S3

        Args:
            pattern (str): pattern to match model files
            dest_dir (str): destination directory
            s3_base_url (str, optional): base url of S3.
        """
        if dest_dir is None:
            dest_dir = str(self.store_dir / "models")
        logger.info(
            f"Downloading models from {s3_base_url}"
            f" with pattern {pattern}"
            f" to {dest_dir}"
        )
        xml_content = download_content(s3_base_url)
        ori_items = parse_s3_xml(xml_content, pattern)
        # filter out existing files
        urls = []
        for item in ori_items:
            target_path = Path(dest_dir) / item["key"]
            expected_size = item["size"]
            if target_path.exists() and (target_path.stat().st_size == expected_size):
                logger.info(f"File exists: {target_path}")
                continue
            if item["key"].endswith("/"):
                continue
            urls.append(f"{s3_base_url}{item['key']}")
        logger.info(f"Founds {len(urls)} files to download.")
        await download_files(
            urls, dest_dir, n_parallel=n_parallel, base_url=s3_base_url
        )

    async def download_models_from_s3(
        self,
        pattern: str,
        dest_dir: T.Optional[str] = None,
        n_parallel: int = 5,
        s3_base_url=S3_MODELS_URL,
    ):
        """Download models from S3"""
        await self.download_from_s3(
            pattern, dest_dir, n_parallel, s3_base_url
        )

    async def download_conda_envs_from_s3(
        self,
        pattern: str,
        dest_dir: T.Optional[str] = None,
        n_parallel: int = 5,
        s3_base_url=S3_CONDA_ENVS_URL,
    ):
        """Download conda envs from S3"""
        await self.download_from_s3(
            pattern, dest_dir, n_parallel, s3_base_url
        )

    def pull_image(self, image_name: str):
        self.container_engine.pull_image(image_name)

    async def launch_s3_server(
            self,
            minio_root_user: T.Optional[str] = None,
            minio_root_password: T.Optional[str] = None):
        """Run a worker server, run in the login node of HPC. """
        engine = self.engine
        self.container_engine.pull_image(S3_IMAGE)

        data_dir = self.store_dir / "s3_data"
        data_dir.mkdir(exist_ok=True, parents=True)
        if minio_root_user is None:
            minio_root_user = "minio"
        if minio_root_password is None:
            minio_root_password = secrets.token_urlsafe(16)

        async def start_s3_server():

            port = PortManager.get_port()
            console_port = PortManager.get_port()
            cmd = self.container_engine.get_command(
                f'server /data --console-address ":{port}" --address ":{console_port}"',  # noqa
                S3_IMAGE,
                ports={
                    port: port,
                    console_port: console_port,
                },
                volumes={data_dir.as_posix(): "/data"},
            )
            env = os.environ.copy()
            env["MINIO_ROOT_USER"] = minio_root_user
            env["MINIO_ROOT_PASSWORD"] = minio_root_password
            job = SubprocessJob(
                cmd, base_class=ProcessJob,
                popen_kwargs={"env": env},
            )
            await engine.submit_async(job)
            await job.wait_until_status("running")
            return job, port, console_port

        job, port, console_port = await start_s3_server()
        return {
            "job_id": job.id,
            "console_port": port,
            "port": console_port,
            "stop": job.cancel
        }

    async def launch_job(self, job: Job):
        """Launch an executor job."""
        await self.engine.submit_async(job)
        return {
            "job_id": job.id,
            "stop": job.cancel
        }

    async def launch_command(
            self,
            cmd: str,
            image_name: T.Optional[str] = None,
            container_kwargs: T.Optional[T.Dict] = None,
            hpc_kwargs: T.Optional[T.Dict[str, str]] = None,
            cmd_kwargs: T.Optional[T.Dict[str, str]] = None,
            ):
        if image_name is not None:
            self.container_engine.pull_image(image_name)
            container_kwargs = container_kwargs or {}
            cmd = self.container_engine.get_command(cmd, image_name, **container_kwargs)
        cmd = self.hpc_manager.get_command(cmd, **(hpc_kwargs or {}))
        job = SubprocessJob(cmd, base_class=ProcessJob, **(cmd_kwargs or {}))
        await self.engine.submit_async(job)
        return {
            "job_id": job.id,
            "stop": job.cancel
        }

    async def launch_server_app(self, server, app_code: str):
        from .constants import IMJOY_APP_TEMPLATE
        controller = await server.get_service("server-apps")
        imjoy_app_code = IMJOY_APP_TEMPLATE.format(app_code=app_code)
        config = await controller.launch(
            source=imjoy_app_code,
            config={"type": "web-python"},
        )
        return config

    async def _get_hypha_server(self, server):
        if isinstance(server, str):
            server = await connect_to_server({"server_url": server})
        return server

    async def launch_ip_record_server(self):
        self._ip_record_flie.parent.mkdir(exist_ok=True, parents=True)
        record_file = self._ip_record_flie.as_posix()

        def run_server(ip, port: int):
            from aiohttp import web
            uuid_to_address = {}

            async def handle(request):
                forwarded_for = request.headers.get('X-Forwarded-For')
                if forwarded_for:
                    client_ip = forwarded_for.split(',')[0].strip()
                else:
                    peername = request.transport.get_extra_info('peername')
                    if peername is not None:
                        client_ip = peername[0]
                    else:
                        client_ip = 'Unknown'

                print(f"Client IP: {client_ip}")
                data = await request.json()
                task_port = int(data.get("port", 0))
                task_uuid = data.get("uuid", "")
                uuid_to_address[task_uuid] = (client_ip, task_port)
                with open(record_file, "w") as f:
                    for _uuid, (_ip, _port) in uuid_to_address.items():
                        f.write(f"{_uuid} {_ip} {_port}\n")
                return web.Response(text=f"Hello, your IP is {client_ip}")

            app = web.Application()
            app.router.add_post('/', handle)
            web.run_app(app, host=ip, port=port)

        job = WebappJob(run_server, ip="0.0.0.0", base_class=ProcessJob)
        self._ip_record_server_job = job
        job_dict = await self.launch_job(job)
        job_dict["port"] = job.port
        await job.wait_until_status("running")
        return job_dict

    async def launch_triton_server(
            self,
            models_dir: T.Optional[str] = None,
            **kwargs):
        """Launch a Triton worker."""
        self.container_engine.pull_image(TRITON_IMAGE)
        task_uuid = str(uuid.uuid4())
        host_ips = [v[1] for v in get_all_ips()]
        if (self._ip_record_server_job is None) or (self._ip_record_server_job.status != "running"):
            await self.launch_ip_record_server()
        if models_dir is None:
            models_dir = (self.store_dir / "models").as_posix()
        Path(models_dir).mkdir(exist_ok=True, parents=True)
        launch_script = LAUNCH_TRITON_SCRIPT.format(
            task_uuid=task_uuid,
            host_ips=repr(host_ips),
            ip_record_server_port=repr(self._ip_record_server_job.port),  # type: ignore
            container_engine_kwargs=repr(self._container_engine_kwargs),
            model_repository=models_dir,
        )
        script_dir = self.store_dir / "tmp"
        script_dir.mkdir(exist_ok=True)
        script_path = script_dir / f"launch_triton_{task_uuid}.py"
        with open(script_path, "w") as f:
            f.write(launch_script)
        run_cmd = f"python {script_path.as_posix()}"
        cmd = self.hpc_manager.get_command(run_cmd, **kwargs)
        job = SubprocessJob(cmd, base_class=ProcessJob)
        job_dict = {
            "job_id": job.id,
            "stop": job.cancel,
        }
        await self.engine.submit_async(job)
        await job.wait_until(lambda j: j.status in ("running", "failed"))
        address_found = False
        while not address_found:  # wait for the address to be recorded
            if self._ip_record_flie.exists():
                with open(self._ip_record_flie, "r") as f:
                    lines = f.readlines()
                    for line in lines:
                        _uuid, _ip, _port = line.strip().split(" ")
                        if _uuid == task_uuid:
                            job_dict["address"] = f"{_ip}:{_port}"
                            address_found = True
                            break
            await asyncio.sleep(1)
        if job.status == "failed":
            raise ValueError("Failed to launch Triton server")
        self._task_uuid_to_job[task_uuid] = job
        return job_dict

    async def launch_bioimageio_backend(
            self,
            models_dir: T.Optional[str] = None,
            hypha_server_url: str = "https://ai.imjoy.io/",
            service_name: str = "triton",
            service_id: T.Optional[str] = None,
            service_config: T.Optional[dict] = None,
            ):
        if models_dir is None:
            models_dir = (self.store_dir / "models").as_posix()
        await self.download_models_from_s3(".*", models_dir)
        await self.launch_ip_record_server()
        triton_job = await self.launch_triton_server(models_dir)
        triton_address = triton_job['address']

        async def get_triton_config(model_name: str, verbose: bool = False):  # noqa
            try:
                res = await get_config(
                    triton_address,
                    model_name=model_name,
                    verbose=verbose,
                )
                return res
            except Exception as e:
                logger.error(f"Error: {e}")
                return {"error": str(e)}

        async def execute_triton(
                inputs: T.Union[T.Any, None] = None,
                model_name: T.Union[str, None] = None,
                cache_config: bool = True,
                **kwargs):
            try:
                res = await execute(
                    inputs=inputs,
                    server_url=triton_address,
                    model_name=model_name,
                    cache_config=cache_config,
                    **kwargs,
                )
                return res
            except Exception as e:
                logger.error(f"Error: {e}")
                return {"error": str(e)}

        server = await self._get_hypha_server(hypha_server_url)
        if service_id is None:
            service_id = f"{service_name}-{secrets.token_urlsafe(8)}"
        logger.info(f"Registering service: {service_name} with id: {service_id}")
        if service_config is None:
            service_config = {"visibility": "public"}
        await server.register_service(
            {
                "name": service_name,
                "id": service_id,
                "config": service_config,
                "get_config": get_triton_config,
                "execute": execute_triton,
            }
        )

        await self.engine.wait_async()

    async def launch_hello_world(self):
        """Detect in which environment, docker/k8s/apptainer"""
        # detect env
        # print env name
        # launch a ubuntu, echo hello world from {ENV_NAME}
        pass


def create_service(config):
    launcher = HyphaLauncher(config)
    return {
        "service_id": "hypha-launcher",
        "pull_image": launcher.pull_image,
        "download_models_from_s3": launcher.download_models_from_s3,
        "download_conda_envs_from_s3": launcher.download_conda_envs_from_s3,
        "launch_s3_server": launcher.launch_s3_server,
        "launch_triton_server": launcher.launch_triton_server,
        "launch_command": launcher.launch_command,
        "stop_job": launcher.stop_job,
        "get_jobs_ids": launcher.get_jobs_ids,
    }



import asyncio
from hypha_launcher.api import HyphaLauncher

import pytest


@pytest.mark.asyncio
async def test_launch_s3():
    hypha_launcher = HyphaLauncher()
    job = await hypha_launcher.launch_s3_server()
    await asyncio.sleep(10)
    await job['stop']()
    await asyncio.sleep(1)


@pytest.mark.asyncio
async def test_launch_triton():
    hypha_launcher = HyphaLauncher()
    ip_job = await hypha_launcher.launch_ip_record_server()
    job = await hypha_launcher.launch_triton_server()
    print(job['address'])
    await asyncio.sleep(15)
    await job['stop']()
    await ip_job['stop']()


@pytest.mark.asyncio
async def test_launch_command():
    hypha_launcher = HyphaLauncher(hpc_manager_kwargs={"hpc_type": "local"})
    port = hypha_launcher.get_free_port()
    job = await hypha_launcher.launch_command(f"python -m http.server {port}")
    await asyncio.sleep(5)
    await job['stop']()
    await asyncio.sleep(1)



import typing as T
from pathlib import Path

from .log import get_logger
from .misc import run_cmd

logger = get_logger()


class ContainerEngine:
    """Container engine abstraction.
    Provides a common interface to container engines,
    such as docker, apptainer, podman, etc.
    """

    supported_engines = ["docker", "apptainer", "podman"]

    def __init__(
        self,
        store_dir: str = "~/.hypha_launcher/containers",
        engine_type: T.Optional[str] = None,
    ):

        self.store_dir = Path(store_dir).expanduser()
        if not self.store_dir.exists():
            self.store_dir.mkdir(parents=True)
        if engine_type is not None:
            self.engine_type = engine_type
        else:
            self.engine_type = self.detect_engine_type()
        self.sif_files: T.Dict[str, Path] = {}

    def detect_engine_type(self):
        for engine in self.supported_engines:
            try:
                run_cmd([engine, "--version"], check=True)
                logger.info(f"Found supported container engine: {engine}")
                return engine
            except FileNotFoundError:
                pass
        else:
            raise RuntimeError(
                "Cannot find supported container engine: " f"{self.supported_engines}"
            )

    @staticmethod
    def process_image_name_for_docker(image_name: str):
        if image_name.startswith("docker://"):
            return image_name[len("docker://"):]
        return image_name

    @staticmethod
    def process_image_name_for_podman(image_name: str):
        return f"{ContainerEngine.process_image_name_for_docker(image_name)}"

    def pull_image(self, image_name: str):
        logger.info(f"Pulling image {image_name}")
        if self.engine_type == "docker":
            image_name = self.process_image_name_for_docker(image_name)
            run_cmd(["docker", "pull", image_name], check=True)
        elif self.engine_type == "apptainer":
            sif_prefix = image_name.split("//")[1].replace("/", "-").replace(":", "_")
            sif_path = self.store_dir / f"{sif_prefix}.sif"
            if not Path(sif_path).exists():
                logger.info(f"Pull and saving image to {sif_path}")
                run_cmd(["apptainer", "pull", str(sif_path), image_name], check=True)
            else:
                logger.info(f"Image exists: {sif_path}")
            self.sif_files[image_name] = sif_path
        elif self.engine_type == "podman":
            image_name = self.process_image_name_for_podman(image_name)
            run_cmd(["podman", "pull", image_name], check=True)
        else:
            raise NotImplementedError

    def get_command(
        self,
        cmd: str,
        image_name: str,
        volumes: T.Optional[dict] = None,
        ports: T.Optional[dict] = None,
        envs: T.Optional[dict] = None,
    ) -> str:
        """Get the command for run process in the container

        Args:
            cmd (str): command to run in the container
            image_name (str): image name
            volumes (dict, optional): volume mapping
                The key is the host path and the value is the container path
            ports (dict, optional): port mapping
                The key is the host port and the value is the container port
            envs (dict, optional): environment variables
                The key is the environment variable name and the value is the value
        """
        # Initialize volume and port mappings as empty strings
        volume_mapping = ""
        port_mapping = ""
        env_options = ""
        if volumes is None:
            volumes = {}
        # If volumes are provided, construct volume mapping options
        for host_path, container_path in volumes.items():
            volume_mapping += f"-v {host_path}:{container_path} "
        # If ports are provided, construct port mapping options
        if ports:
            for host_port, container_port in ports.items():
                port_mapping += f"-p {host_port}:{container_port} "
        # If environment variables are provided, construct environment variable options
        if envs:
            for env_name, env_value in envs.items():
                env_options += f"-e {env_name}={env_value} "
        # Construct the command based on the engine type
        if self.engine_type == "docker":
            image_name = self.process_image_name_for_docker(image_name)
            return f"docker run --rm {env_options} {volume_mapping} {port_mapping} {image_name} {cmd}"
        elif self.engine_type == "apptainer":
            # add tmp dir mapping when using apptainer
            host_tmp_dir = self.store_dir / "apptainer_tmp"
            host_tmp_dir.mkdir(exist_ok=True)
            volumes[host_tmp_dir.as_posix()] = "/tmp"
            # Note: Apptainer (formerly Singularity) has different options
            sif_path = self.sif_files[image_name]
            # For Apptainer, bind options are used for volume mapping
            bind_option = ""
            if volumes:
                binds = [f"{host}:{container}" for host, container in volumes.items()]
                bind_option = f"--bind {','.join(binds)} "
            if envs:
                env_options = " ".join([f"--env {k}={v}" for k, v in envs.items()])
            return f"apptainer run --contain {env_options} {bind_option} {sif_path} {cmd}"
        elif self.engine_type == "podman":
            image_name = self.process_image_name_for_docker(image_name)
            return f"podman run {env_options} {volume_mapping} {port_mapping} {image_name} {cmd}"
        else:
            raise NotImplementedError


if __name__ == "__main__":
    import fire

    fire.Fire(ContainerEngine)



import typing as T
import os
import subprocess as subp
import socket

import psutil

from .log import get_logger


logger = get_logger()


def run_cmd(cmd: T.Union[str, T.List[str]], check: bool = True, **kwargs):
    if isinstance(cmd, list):
        logger.info(f"Running command: {' '.join(cmd)}")
        subp.run(cmd, check=check, **kwargs)
    else:
        logger.info(f"Running command: {cmd}")
        subp.run(cmd, shell=True, check=check, **kwargs)


def get_all_ips() -> T.List[T.Tuple[str, str]]:
    ip_info = []
    for interface, addrs in psutil.net_if_addrs().items():
        for addr in addrs:
            if addr.family == socket.AF_INET:  # IPv4
                ip_info.append((interface, addr.address))
    return ip_info


def check_ip_port(ip, port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(10)
    result = sock.connect_ex((ip, port))
    sock.close()
    return result == 0


def detect_runtime_environment():
    # Check for Docker and Podman through /proc/self/cgroup
    try:
        with open('/proc/self/cgroup', 'rt') as ifh:
            cgroup_contents = ifh.read()
            if 'docker' in cgroup_contents or '/docker/' in cgroup_contents:
                return 'Docker'
            elif 'podman' in cgroup_contents or '/libpod/' in cgroup_contents:
                return 'Podman'
    except FileNotFoundError:
        pass  # /proc/self/cgroup does not exist, not in a container, or not allowed to read

    # Check for Kubernetes by looking for specific environment variables
    if os.getenv('KUBERNETES_SERVICE_HOST'):
        return 'Kubernetes'

    # Check for Apptainer/Singularity by looking for environment variables set by it
    if os.getenv('SINGULARITY_CONTAINER') or os.getenv('APPTAINER_CONTAINER'):
        return 'Apptainer/Singularity'

    # Check for .dockerenv file
    if os.path.exists('/.dockerenv'):
        return 'Docker'

    # No indicators found
    return 'Unknown'



import os
import typing as T
from .log import get_logger
from .misc import run_cmd

logger = get_logger()


def detect_hpc_type() -> str:
    try:
        run_cmd(["sinfo"], check=True)
        return "slurm"
    except FileNotFoundError:
        pass
    try:
        run_cmd(["qstat"], check=True)
        return "pbs"
    except FileNotFoundError:
        pass
    return "local"


class HPCManger:
    def __init__(self, hpc_type: T.Optional[str] = None):
        if hpc_type is None:
            hpc_type = detect_hpc_type()
        self.hpc_type = hpc_type

    def get_command(self, cmd: str, **attrs) -> str:
        if self.hpc_type == "slurm":
            return self.get_slurm_command(cmd, **attrs)
        elif self.hpc_type == "local":
            return cmd
        else:
            raise NotImplementedError(f"Unsupported HPC type: {self.hpc_type}")

    def get_slurm_command(
            self,
            cmd: str,
            account: T.Optional[str] = None,
            time: T.Optional[str] = None,
            gpus_per_node: T.Optional[str] = None,
            additonal_options: T.Optional[str] = None,
            ) -> str:
        if account is None:
            account = os.environ.get("SLURM_ACCOUNT")
            if account is None:
                raise ValueError("SLURM_ACCOUNT is not set.")
        if time is None:
            time = os.environ.get("SLURM_TIME")
        if gpus_per_node is None:
            gpus_per_node = os.environ.get("SLURM_GPUS_PER_NODE")
        if additonal_options is None:
            additonal_options = os.environ.get("SLURM_ADDITIONAL_OPTIONS")

        options_str = f"--account={account} "
        if time is not None:
            options_str += f"--time={time} "
        if gpus_per_node is not None:
            options_str += f"--gpus-per-node={gpus_per_node} "
        if additonal_options is not None:
            options_str += additonal_options
        new_cmd = f"srun {options_str} {cmd}"
        return new_cmd
